@@ -45,7 +45,7 @@ class RequestUtil:
                  auth_browser_success_msg: str,
                  auth_browser_fail_msg: str,
                  retrieve_token: T.Callable[[TokenType], T.Optional[dict]],
-                 persist_token: T.Callable[[TokenType, dict], None]):
+                 persist_token: T.Callable[[TokenType, T.Optional[dict]], None]):
 
         self.api_url = api_url
         self.idp_url = idp_url
@@ -106,6 +106,10 @@ class RequestUtil:
             logging.debug("No refresh token found")
             return False
 
+        if time.time() > refresh_token.expires_at - self.auth_token_min_valid_sec:
+            logging.debug("Refresh token expired")
+            return False
+
         token_req_body = {
             'grant_type': "refresh_token",
             'refresh_token': refresh_token.token,
@@ -228,8 +232,8 @@ class RequestUtil:
         if token_dict is not None:
             return StorableToken(**token_dict)
 
-    def set_stored_token(self, token_type: TokenType, token: StorableToken):
-        self.persist_token(token_type, dataclasses.asdict(token))
+    def set_stored_token(self, token_type: TokenType, token: T.Optional[StorableToken]):
+        self.persist_token(token_type, None if token is None else dataclasses.asdict(token))
 
 
 class Common:
@@ -321,6 +325,35 @@ class Teacher:
         path = "/teacher/courses"
         return self.request_util.simple_get_request(path, data.TeacherCourseResp)
 
+    def get_course_participants(self, course_id: str, role: data.ParticipantRole = data.ParticipantRole.ALL,
+                                limit: int = 1_000_000, offset: int = 0) -> data.TeacherCourseParticipantsResp:
+        """
+        TODO
+        """
+        logging.debug(f"Get participants on course {course_id} with role {role} (offset: {offset}, limit: {limit})")
+        path = f"/courses/{course_id}/participants?role={role.value}&offset={offset}&limit={limit}"
+        return self.request_util.simple_get_request(path, data.TeacherCourseParticipantsResp)
+
+    def get_course_exercises(self, course_id: str) -> data.TeacherCourseExercisesResp:
+        """
+        TODO
+        """
+        logging.debug(f"Get teacher exercises on course {course_id}")
+        path = f"/teacher/courses/{course_id}/exercises"
+        return self.request_util.simple_get_request(path, data.TeacherCourseExercisesResp)
+
+    def get_course_exercise_submissions_student(self, course_id: str, course_exercise_id: str, student_id: str,
+                                                limit: int = 1_000_000,
+                                                offset: int = 0) -> data.TeacherCourseExerciseSubmissionsStudentResp:
+        """
+        TODO
+        """
+        logging.debug(f"Get submissions to course exercise {course_exercise_id} on course {course_id} by "
+                      f"student {student_id} (offset: {offset}, limit: {limit})")
+        path = f"/teacher/courses/{course_id}/exercises/{course_exercise_id}/submissions/all/students/{student_id}" \
+               f"?offset={offset}&limit={limit}"
+        return self.request_util.simple_get_request(path, data.TeacherCourseExerciseSubmissionsStudentResp)
+
 
 # TODO: hide private fields/methods
 # TODO: should use TokenStorer type/class instead of functions?
@@ -352,7 +385,10 @@ class Ez:
             return local_token_store[token_type] if token_type in local_token_store else None
 
         def in_memory_persist_token(token_type, token):
-            local_token_store[token_type] = token
+            if token is None:
+                local_token_store.pop(token_type, None)
+            else:
+                local_token_store[token_type] = token
 
         # ======
 
@@ -392,7 +428,7 @@ class Ez:
     def start_auth_in_browser(self):
         self.util.start_auth_in_browser()
 
-    def is_auth_in_progress(self, timeout_sec: int = 0) -> bool:
+    def is_auth_in_progress(self, timeout_sec: T.Optional[int] = 0) -> bool:
         thread = self.util.auth_server_thread
         if thread is None:
             return False
@@ -400,6 +436,10 @@ class Ez:
             thread.join(timeout_sec)
             return thread.is_alive()
 
+    def await_is_auth_completed(self, timeout_sec: T.Optional[int] = None) -> bool:
+        self.is_auth_in_progress(timeout_sec)
+        return not self.is_auth_required()
+
     def is_auth_required(self) -> bool:
         try:
             self.util.get_valid_access_token()
@@ -423,5 +463,9 @@ class Ez:
                 logging.warning(f'Got exception {repr(e)}')
 
     def logout_in_browser(self):
-        url = f"/auth/realms/master/protocol/openid-connect/logout?redirect_uri=https%3A%2F%2F"
-        webbrowser.open(self.util.idp_url + url + self.util.idp_client_name)
+        self.util.set_stored_token(TokenType.ACCESS, None)
+        self.util.set_stored_token(TokenType.REFRESH, None)
+        webbrowser.open(
+            f"{self.util.idp_url}/auth/realms/master/protocol/openid-connect/logout"
+            f"?redirect_uri=https%3A%2F%2F{self.util.idp_client_name}"
+        )
